/*
 * *********************************************************************************************************************
 *
 *          !!
 *        .F88X
 *        X8888Y
 *      .}888888N;
 *        i888888N;        .:!              .I$WI:
 *          R888888I      .'N88~            i8}+8Y&8"l8i$8>8W~'>W8}8]KW+8IIN"8&
 *          .R888888I    .;N8888~          .X8'  "8I.!,/8"  !%NY8`"8I8~~8>,88I
 *            +888888N;  .8888888Y                                  "&&8Y.}8,
 *            ./888888N;  .R888888Y        .'}~    .>}'.`+>  i}!    "i'  +/'  .'i~  !11,.:">,  .~]!  .i}i
 *              ~888888%:  .I888888l      .]88~`1/iY88Ii+1'.R$8$8]"888888888>  Y8$  W8E  X8E  W8888'188Il}Y88$*
 *              18888888    E8888881    .]W%8$`R8X'&8%++N8i,8N%N8+l8%`  .}8N:.R$RE%N88N%N$K$R  188,FE$8%~Y88I
 *            .E888888I  .i8888888'      .:$8I;88+`E8R:/8N,.>881.`$8E/1/]N8X.Y8N`"KF&&FK!'88*."88K./$88%RN888+~
 *            8888888I  .,N888888~        ~88i"8W,!N8*.I88.}888%F,i$88"F88"  888:E8X.>88!i88>`888*.}Fl1]*}1YKi'
 *          i888888N'      I888Y          ]88;/EX*IFKFK88X  K8R  .l8W  88Y  ~88}'88E&%8W.X8N``]88!.$8K  .:W8I
 *        .i888888N;        I8Y          .&8$  .X88!  i881.:%888>I88  ;88]  +88+.';;;;:.Y88X  18N.,88l  .+88/
 *      .:R888888I
 *      .&888888I                                          Copyright (c) 2009-2020.  博锐尚格科技股份有限公司
 *        ~8888'
 *        .!88~                                                                     All rights reserved.
 *
 * *********************************************************************************************************************
 */

package com.persagy.mysql

import com.persagy.base.extensions.isBaseType
import com.persagy.base.extensions.toJson
import com.persagy.base.extensions.toJsonAll
import com.persagy.database.SAbstractFunction
import com.persagy.database.SFilter
import com.persagy.database.enums.SDatabaseType
import com.persagy.database.enums.SOps
import com.persagy.mybatis.SAbstractDriver
import com.persagy.mybatis.SSqlProvider
import org.apache.ibatis.jdbc.SQL
import org.slf4j.LoggerFactory

/**
 * MySQL 驱动
 *
 * @author 庞利祥 <sybotan@126.com>
 */
class SMySqlDriver : SAbstractDriver() {
    /** 静态对象 */
    companion object {
        /** 日志记录器 */
        private val logger = LoggerFactory.getLogger(SMySqlDriver::class.java)
    }

    /**
     * 在数据库中插入实体
     *
     * @param argsMap         注入参数
     * @return 插入 SQL 语句
     */
    @Suppress("UNCHECKED_CAST")
    override fun insert(argsMap: Map<String, Any?>): String {
        val entityClass = argsMap["entityClass"] as Class<*>
        val entity = argsMap["entity"] as HashMap<String, Any?>
        val sql = object: SQL() {
            init {
                INSERT_INTO(SSqlProvider.driver.tableName(entityClass))
                for ((key, value) in entity) {
                    if (value != null) {
                        if (value is Map<*,*>) {
                            VALUES(SSqlProvider.driver.escName(key), "'${value.toJsonAll().replace("'", "''")}'")
                        } else {
                            VALUES(SSqlProvider.driver.escName(key), SSqlProvider.driver.entityValue(entity, key))
                        }
//                        if (value::class.java.isPrimitive) {
//                            VALUES(driver.escName(key), driver.entityValue(entity, key))
//                        } else {
//                            VALUES(driver.escName(key), "'${value.toJsonAll()}'")
//                        }
                    }
                }
            }
        }.toString()

        logger.debug("INSERT SQL= $sql")

        return sql
    }

    /**
     * 在数据库中替换实体
     *
     * @param argsMap         注入参数
     *
     * @return 替换 SQL 语句
     */
    override fun replace(argsMap: Map<String, Any?>): String {
        val insertSql = ""
        val sql = insertSql.replace("INSERT INTO", "REPLACE INTO")
        logger.debug("REPLACE SQL= $sql")
        return sql
    }

    /**
     * 根据条件删除
     *
     * @param argsMap         注入参数
     * @return 删除 SQL 语句
     */
    @Suppress("UNCHECKED_CAST")
    override fun delete(argsMap: Map<String, Any?>): String {
        val entityClass = argsMap["entityClass"] as Class<*>
        /** 从入参中获得过虑条件 */
        val filterArgs = argsMap["filterArgs"] as List<SFilter>
        /** 转换后的where语句参数 */
        val whereArgs = argsMap["whereArgs"] as HashMap<String, Any>
        val clause = buildWhereArgList(entityClass, filterArgs, whereArgs)

        val sql = object: SQL() {
            init {
                DELETE_FROM(SSqlProvider.driver.tableName(entityClass))
                if (clause.isNotEmpty()) {
                    WHERE(clause)
                }
            }
        }.toString()

        logger.debug("SQL= $sql")
        return sql
    }

    /**
     * 删除所有记录
     *
     * @param argsMap         注入参数
     * @return  删除SQL语句
     */
    @Suppress("UNCHECKED_CAST")
    override fun deleteAll(argsMap: Map<String, Any?>): String {
        val entityClass = argsMap["entityClass"] as Class<*>
        val sql = object: SQL() {
            init {
                DELETE_FROM(SSqlProvider.driver.tableName(entityClass))
            }
        }.toString()

        logger.debug("SQL= $sql")
        return sql
    }

    /**
     * 获得表名，取实体类@Table注解的名称，如果没有注解，则取实体类类名。
     *
     * @param clz         mapper 类类型
     * @return 表名
     */
    override fun tableName(clz: Class<*>): String {
        val tableAnno = clz.getAnnotation(javax.persistence.Table::class.java)
        /** 如果有@Table注解，则取注解的表名；否则使用类名做为表名 */
        val tableName = tableAnno?.name ?: clz.simpleName
        val name = tableName.replace(".", "`.`")
        return "`$name`"
    }

    /**
     * 将 name 加上引用符号
     *
     * @param name    名称
     * @return 加上引用符号的名称
     */
    override fun escName(name: String): String {
        return "`$name`"
    }

    /**
     * 实体类的值，如果不是基本类型，则转换为json
     *
     * @param entity      实体类
     * @param name        名称
     * @return 加上引用符号的名称
     */
    override fun entityValue(entity: HashMap<String, Any?>, name: String): String {
        val value = entity[name] ?: return "null"
        if (value.isBaseType()) {
            return "#{entity.$name}"
        }
        return "'${value.toJson()}'"
    }

    /**
     * 生成 limit 语句
     *
     * @param offset      偏移位置
     * @param count       取得的记录数
     * @return LIMIT 子句
     */
    override fun limit(offset: Int?, count: Int?): String {
        return " LIMIT ${offset ?: 0}, ${count ?: 1}"
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    // 函数
    /**
     * 处理函数支持
     *
     * @param func        函数对象
     * @param colName     字段名
     * @return 生成的 SQL 语句
     */
    override fun processFunction(func: SAbstractFunction, colName: String): String {
        func.colName = colName
        return func.toSql(SDatabaseType.MySQL)
    }

    override fun updateNullItem(nullItem: String): String {
        TODO("not implemented") //To change body of created functions use File | Settings | File Templates.
    }
//----------------------------------------------------------------------------------------------------------------------
    /**
     * 转换where字句参数
     *
     * @param argName     参数名
     * @param argValue    参数值
     * @param ops         运算符
     * @return 转换后的 SQL
     */
    override fun whereArg(argName: String, argValue: String, ops: SOps): String {
//        var name = obj as String
//
//        val colName: String
//        val argValue: String
//        if (path == null) {
//            colName = "`$name`"
//            argValue = "#{whereArgs.$argName}"
//        } else {
//            colName = "JSON_EXTRACT(`$name`, '\$.$path')"
//            argValue = "#{whereArgs.$argName}"
//        }

        when (ops) {
            SOps.EQ -> {     /** 等于 */
                return "$argName = $argValue"
            }
            SOps.NE -> {     /** 不等于 */
                return "$argName <> $argValue"
            }
            SOps.GT -> {     /** 大于 */
                return "$argName > $argValue"
            }
            SOps.GTE -> {    /** 大于等于 */
                return "$argName >= $argValue"
            }
            SOps.LT -> {     /** 小于 */
                return "$argName < $argValue"
            }
            SOps.LTE -> {    /** 小于等行 */
                return "$argName <= $argValue"
            }
            SOps.CONTAIN -> {    /** 包含 */
                return "$argName LIKE CONCAT(\"%\" , $argValue, \"%\")"
            }
            SOps.STARTWITH -> {    /** 以指定字符串开始 */
                return "$argName LIKE CONCAT($argValue, \"%\")"
            }
            SOps.ENDWITH -> {    /** 以指定字符串结束 */
                return "$argName LIKE CONCAT(\"%\" , $argValue)"
            }
            SOps.ISNULL -> {     /** 为空 */
                return "ISNULL($argName)"
            }
            else -> {
            }
        }
        return ""
    }

    /**
     * 转换参数名 SQL 语句
     *
     * @param name        字段名
     * @param path        json 路径，如果不是 json 对象，则该值为空
     * @return 转换后的 SQL
     */
    override fun argName(name: String, path: String?): String {
        return if (path == null) {
            "`$name`"
        } else {
            "JSON_EXTRACT(`$name`, '\$.$path')"
        }
    }

    /**
     * 转换参数值 SQL 语句
     *
     * @param argName     参数名
     * @param isJson      是否 Json 对象
     * @return 转换后的 SQL
     */
    override fun argValue(argName: String, isJson: Boolean): String {
        return if (isJson) {
            "#{whereArgs.$argName}"
        } else {
            "#{whereArgs.$argName}"
        }
    }

    /**
     * 转换 IN 条件表达式
     *
     * @param name        字段名
     * @param path        json 路径，如果字段不为 json，则 path 为 null
     * @param inArg       in 操作参数
     * @return 转换后的 SQL
     */
    override fun inArg(name: String, path: String?, inArg: String): String {
        return if (path == null) {
            "`$name` IN ($inArg)"
        } else {
            "JSON_EXTRACT(`$name`,'\$.$path') IN ($inArg)"
        }
    }

    /**
     * 将名称及 json 路径转换为 SQL
     *
     * @param name        字段名
     * @param path        json 路径，如果字段不为 json，则 path 为 null
     * @return 转换后的 SQL
     */
    override fun columnName(name: String, path: String?): String {
        return if (path == null) {
            "`$name`"
        } else {
            "JSON_EXTRACT(`$name`, '\$.$path')"
        }
    }

    /**
     * 将参数值转换为 SQL 语句
     *
     * @param value       值
     * @param isJson      值是否转换为 json 格式
     * @return 转换后的 SQL
     */
    override fun argValueToJson(value: String, isJson: Boolean): String {
        return value
    }

    /**
     * 更新 Json 类型字段
     *
     * @param name        字段名
     * @param value       值
     * @return 更新 SQL 字符串
     */
    override fun updateJson(name: String, value: Any): String {
        return "JSON_MERGE(IFNULL(`$name`,'{}'), '${value.toJsonAll()}')"
    }

}
